from __future__ import annotations

import warnings
from collections import abc as cabc
from collections import defaultdict
from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager
from functools import reduce
from typing import (
    TYPE_CHECKING,
    Literal,
    TypeVar,
)

import anndata as ad
import networkx as nx
import pandas as pd

from treedata._utils import subset_tree

if TYPE_CHECKING:
    from anndata import AnnData
    from anndata.raw import Raw

    from treedata._core.treedata import TreeData


OneDIdx = Sequence[int] | Sequence[bool] | slice
TwoDIdx = tuple[OneDIdx, OneDIdx]

I = TypeVar("I", OneDIdx, TwoDIdx, covariant=True)


class AxisTreesBase(cabc.MutableMapping):
    """Mapping of key to nx.DiGraph aligned to an axis of parent TreeData."""

    def __repr__(self):
        return f"{type(self).__name__} with keys: {', '.join(self.keys())}"

    def _ipython_key_completions_(self) -> list[str]:
        return list(self.keys())

    def _validate_tree(self, tree: nx.DiGraph, key: str) -> nx.DiGraph:
        # Check value type
        if not isinstance(tree, nx.DiGraph):
            raise ValueError(f"Value for key {key} must be a nx.DiGraph")
        # Empty tree
        if tree.number_of_nodes() == 0:
            return tree, set()
        # Check tree
        if tree.number_of_nodes() != tree.number_of_edges() + 1:
            raise ValueError(f"Value for key {key} must be a tree")
        root_count = 0
        leaves = set()
        for node in tree.nodes:
            if tree.in_degree(node) == 0:
                root_count += 1
                if tree.out_degree(node) == 0:
                    raise ValueError(f"Value for key {key} must be fully connected")
            elif tree.in_degree(node) > 1:
                raise ValueError(f"Value for key {key} must be a tree")
            if tree.out_degree(node) == 0:
                leaves.add(node)
        if root_count != 1:
            raise ValueError(f"Value for key {key} must be a tree")
        # Check alignment
        if not leaves.issubset(self.dim_names):
            raise ValueError(f"Leaf names in must be in {self.dim}_names")
        # Check overlap
        if not self.parent.allow_overlap:
            if key in self._tree_to_leaf:
                new_leaves = leaves.difference(self._tree_to_leaf[key])
            else:
                new_leaves = leaves
            if new_leaves.intersection(self._leaf_to_tree.keys()):
                raise ValueError(
                    "Leaf names overlap with leaf names of other trees. Set `allow_overlap=True` to allow this."
                )
        return tree, leaves

    def _update_tree_labels(self):
        if self.parent._tree_label is not None:
            if self.parent.allow_overlap:
                mapping = {k: ",".join(map(str, v)) for k, v in self._leaf_to_tree.items()}
            else:
                mapping = {k: v[0] for k, v in self._leaf_to_tree.items()}
            getattr(self.parent, self.dim)[self.parent._tree_label] = getattr(self.parent, f"{self.dim}_names").map(
                mapping
            )

    def _check_uniqueness(self):
        names = "Observation" if self.dim == "obs" else "Variable"
        if not getattr(self.parent, self.dim).index.is_unique:
            warnings.warn(
                f"{names} names must be unique to store a tree. Calling `.{self.dim}_names_make_unique` to make them unique.",
                stacklevel=2,
            )
            getattr(self.parent, self.dim).index = ad.utils.make_index_unique(getattr(self.parent, self.dim).index)

    def copy(self):
        d = AxisTrees(self.parent, self._axis)
        for k, v in self.items():
            d[k] = v.copy()
        return d

    def _view(self, parent: TreeData, subset_idx: I):
        """Returns a subset copy-on-write view of the object."""
        return AxisTreesView(self, parent, subset_idx)

    @property
    def parent(self) -> AnnData | Raw:
        return self._parent

    @property
    def attrname(self) -> str:
        return f"{self.dim}t"

    @property
    def axes(self) -> tuple[Literal[0, 1]]:
        """Axes of the parent this is aligned to"""
        return (self._axis,)

    @property
    def dim(self) -> str:
        """Name of the dimension this aligned to."""
        return self._dimnames[self._axis]

    @property
    def dim_names(self) -> pd.Index:
        return (self.parent.obs_names, self.parent.var_names)[self._axis]


class AxisTrees(AxisTreesBase):
    def __init__(
        self,
        parent: TreeData,
        axis: int,
        vals: Mapping | None = None,
    ):
        self._parent = parent
        self._dimnames = ("obs", "var")
        if axis not in (0, 1):
            raise ValueError()
        self._axis = axis
        self._data = {}
        self._tree_to_leaf = defaultdict(set)
        self._leaf_to_tree = defaultdict(list)
        if vals is not None:
            self.update(vals)

    def __getitem__(self, key: str) -> nx.DiGraph:
        return nx.graphviews.generic_graph_view(self._data[key])

    def __setitem__(self, key: str, value: nx.DiGraph):
        self._check_uniqueness()
        value, leaves = self._validate_tree(value, key)

        for leaf in leaves:
            self._leaf_to_tree[leaf].append(key)
        self._tree_to_leaf[key] = leaves

        if not self.parent.is_view:
            self._update_tree_labels()

        self._data[key] = value

    def __delitem__(self, key: str):
        for leaf in self._tree_to_leaf[key]:
            self._leaf_to_tree[leaf].remove(key)
            if not self._leaf_to_tree[leaf]:
                del self._leaf_to_tree[leaf]
        del self._tree_to_leaf[key]

        self._update_tree_labels()

        del self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)


class AxisTreesView(AxisTreesBase):
    def __init__(
        self,
        parent_mapping: AxisTreesBase,
        parent_view: TreeData,
        subset_idx: OneDIdx,
    ):
        self.parent_mapping = parent_mapping
        self._parent = parent_view
        self._dimnames = ("obs", "var")
        self.subset_idx = subset_idx
        self._axis = parent_mapping._axis
        self._tree_to_leaf = parent_mapping._tree_to_leaf
        self._leaf_to_tree = parent_mapping._leaf_to_tree

    def __getitem__(self, key: str) -> nx.DiGraph:
        leaves = self.parent_mapping._tree_to_leaf[key]
        subset_leaves = leaves.intersection(self.dim_names.values)
        return subset_tree(self.parent_mapping[key], subset_leaves, asview=True)

    def __setitem__(self, key: str, value: nx.DiGraph):
        self._check_uniqueness()
        value, _ = self._validate_tree(value, key)  # Validate before mutating
        warnings.warn(
            f"Setting element `.{self.attrname}['{key}']` of view, initializing view as actual.", stacklevel=2
        )
        with view_update(self.parent, self.attrname, ()) as new_mapping:
            new_mapping[key] = value

    def __delitem__(self, key: str):
        if key not in self:
            raise KeyError(
                "'{key!r}' not found in view of {self.attrname}"
            )  # Make sure it exists before bothering with a copy
        warnings.warn(
            f"Removing element `.{self.attrname}['{key}']` of view, initializing view as actual.", stacklevel=2
        )
        with view_update(self.parent, self.attrname, ()) as new_mapping:
            del new_mapping[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self.parent_mapping)

    def __contains__(self, key: str) -> bool:
        return key in self.parent_mapping

    def __len__(self) -> int:
        return len(self.parent_mapping)


@contextmanager
def view_update(tdata_view: TreeData, attr_name: str, keys: tuple[str, ...]):
    """Context manager for updating a view of an TreeData object.

    Contains logic for "actualizing" a view. Yields the object to be modified in-place.

    Parameters
    ----------
    adata_view
        A view of an AnnData
    attr_name
        Name of the attribute being updated
    keys
        Keys to the attribute being updated

    Yields
    ------
    `adata.attr[key1][key2][keyn]...`
    """
    new = tdata_view.copy()
    attr = getattr(new, attr_name)
    container = reduce(lambda d, k: d[k], keys, attr)
    yield container
    tdata_view._init_as_actual(new)



from __future__ import annotations

import json
from collections.abc import MutableMapping, Sequence
from pathlib import Path
from typing import (
    Literal,
)

import anndata as ad
import h5py
import zarr
from scipy import sparse

from treedata._core.aligned_mapping import AxisTrees
from treedata._core.treedata import TreeData
from treedata._utils import dict_to_digraph


def _tdata_from_adata(tdata, treedata_attrs=None) -> TreeData:
    """Create a TreeData object parsing attribute from AnnData uns field."""
    tdata.__class__ = TreeData
    if treedata_attrs is not None:
        tdata._tree_label = treedata_attrs["label"] if "label" in treedata_attrs.keys() else None
        tdata._allow_overlap = bool(treedata_attrs["allow_overlap"])
        tdata._obst = AxisTrees(tdata, 0, vals={k: dict_to_digraph(v) for k, v in treedata_attrs["obst"].items()})
        tdata._vart = AxisTrees(tdata, 1, vals={k: dict_to_digraph(v) for k, v in treedata_attrs["vart"].items()})
    else:
        tdata._tree_label = None
        tdata._allow_overlap = False
        tdata._obst = AxisTrees(tdata, 0)
        tdata._vart = AxisTrees(tdata, 1)
    return tdata


def read_h5ad(
    filename: str | Path = None,
    backed: Literal["r", "r+"] | bool | None = None,
    *,
    as_sparse: Sequence[str] = (),
    as_sparse_fmt: type[sparse.spmatrix] = sparse.csr_matrix,
    chunk_size: int = 6000,
) -> TreeData:
    """Read `.h5ad`-formatted hdf5 file.

    Parameters
    ----------
    filename
        File name of data file.
    backed
        If `'r'`, load :class:`~anndata.TreeData` in `backed` mode
        instead of fully loading it into memory (`memory` mode).
        If you want to modify backed attributes of the TreeData object,
        you need to choose `'r+'`.
    as_sparse
        If an array was saved as dense, passing its name here will read it as
        a sparse_matrix, by chunk of size `chunk_size`.
    as_sparse_fmt
        Sparse format class to read elements from `as_sparse` in as.
    chunk_size
        Used only when loading sparse dataset that is stored as dense.
        Loading iterates through chunks of the dataset of this row size
        until it reads the whole dataset.
        Higher size means higher memory consumption and higher (to a point)
        loading speed.
    """
    adata = ad.read_h5ad(
        filename,
        backed=backed,
        as_sparse=as_sparse,
        as_sparse_fmt=as_sparse_fmt,
        chunk_size=chunk_size,
    )
    with h5py.File(filename, "r") as f:
        if "raw.treedata" in f:
            treedata_attrs = json.loads(f["raw.treedata"][()])
        else:
            treedata_attrs = None
    tdata = _tdata_from_adata(adata, treedata_attrs)

    return tdata


def read_zarr(store: str | Path | MutableMapping | zarr.Group) -> TreeData:
    """Read from a hierarchical Zarr array store.

    Parameters
    ----------
    store
        The filename, a :class:`~typing.MutableMapping`, or a Zarr storage class.
    """
    adata = ad.read_zarr(store)

    with zarr.open(store, mode="r") as f:
        if "raw.treedata" in f:
            treedata_attrs = json.loads(f["raw.treedata"][()])
        else:
            treedata_attrs = None
    tdata = _tdata_from_adata(adata, treedata_attrs)

    return tdata



import anndata as ad
import networkx as nx
import numpy as np
import pandas as pd
import pytest

import treedata as td


@pytest.fixture
def X():
    yield np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])


@pytest.fixture
def adata(X):
    yield ad.AnnData(X)


@pytest.fixture
def tree():
    tree = nx.DiGraph()
    tree.add_edges_from([("root", "0"), ("root", "1")])
    yield tree


def check_graph_equality(g1, g2):
    assert nx.is_isomorphic(g1, g2, node_match=lambda n1, n2: n1 == n2, edge_match=lambda e1, e2: e1 == e2)


def test_creation(X, adata, tree):
    # Test creation with np array
    tdata = td.TreeData(X, obst={"tree": tree}, vart={"tree": tree}, label=None)
    print(type(tdata))
    check_graph_equality(tdata.obst["tree"], tree)
    check_graph_equality(tdata.vart["tree"], tree)
    # Test creation with anndata
    tdata = td.TreeData(adata)
    assert tdata.X is adata.X


@pytest.mark.parametrize("axis", [0, 1])
def test_attributes(X, tree, axis):
    dim = ["obs", "var"][axis]
    tdata = td.TreeData(X, obst={"tree": tree}, vart={"tree": tree}, label=None)
    assert getattr(tdata, f"{dim}t").axes == (axis,)
    assert getattr(tdata, f"{dim}t").attrname == (f"{dim}t")
    assert getattr(tdata, f"{dim}t").dim == dim
    assert getattr(tdata, f"{dim}t").parent is tdata
    assert list(getattr(tdata, f"{dim}t").dim_names) == ["0", "1", "2"]
    assert tdata.allow_overlap is False
    assert tdata.label is None


@pytest.mark.parametrize("dim", ["obs", "var"])
def test_tree_set(X, tree, dim):
    tdata = td.TreeData(X)
    setattr(tdata, f"{dim}t", {"tree": tree})
    check_graph_equality(getattr(tdata, f"{dim}t")["tree"], tree)


@pytest.mark.parametrize("dim", ["obs", "var"])
def test_tree_del(X, tree, dim):
    tdata = td.TreeData(X, obst={"tree": tree}, vart={"tree": tree}, label=None)
    del getattr(tdata, f"{dim}t")["tree"]
    assert getattr(tdata, f"{dim}t") == {}


@pytest.mark.parametrize("dim", ["obs", "var"])
def test_tree_contains(X, tree, dim):
    tdata = td.TreeData(X, obst={"tree": tree}, vart={"tree": tree}, label=None)
    assert "tree" in getattr(tdata, f"{dim}t")
    assert "not_tree" not in getattr(tdata, f"{dim}t")


@pytest.mark.filterwarnings
@pytest.mark.parametrize("dim", ["obs", "var"])
def test_tree_label(X, tree, dim):
    # Test tree label
    second_tree = nx.DiGraph()
    second_tree.add_edges_from([("root", "2")])
    tdata = td.TreeData(X, obst={"0": tree, "1": second_tree}, vart={"0": tree, "1": second_tree}, label="tree")
    assert getattr(tdata, dim)["tree"].tolist() == ["0", "0", "1"]
    # Test tree label with overlap
    tdata = td.TreeData(X, obst={"0": tree, "1": tree}, label="tree", vart={"0": tree, "1": tree}, allow_overlap=True)
    assert getattr(tdata, dim).loc["0", "tree"] == "0,1"
    # Test label already present warning
    df = pd.DataFrame({"tree": ["bad", "bad", "bad"]})
    with pytest.warns(UserWarning):
        tdata = td.TreeData(X, label="tree", obs=df, var=df)


def test_tree_overlap(X, tree):
    second_tree = nx.DiGraph()
    second_tree.add_edges_from([("root", "0"), ("root", "1")])
    # Test overlap not allowed
    with pytest.raises(ValueError):
        tdata = td.TreeData(X, obst={"0": tree, "1": second_tree}, allow_overlap=False)
    # Test overlap allowed
    tdata = td.TreeData(X, obst={"0": tree, "1": second_tree}, allow_overlap=True)
    check_graph_equality(tdata.obst["0"], tree)
    check_graph_equality(tdata.obst["1"], second_tree)


def test_repr(X, tree):
    tdata = td.TreeData(X, obst={"tree": tree}, vart={"tree": tree}, label=None)
    # TreeData
    expected_repr = f"TreeData object with n_obs × n_vars = {X.shape[0]} × {X.shape[1]}"
    expected_repr += "\n    obst: 'tree'"
    expected_repr += "\n    vart: 'tree'"
    assert repr(tdata) == expected_repr
    # AxisTrees
    expected_repr = "AxisTrees with keys: tree"
    assert repr(tdata.obst) == expected_repr


def test_mutability(X, tree):
    tdata = td.TreeData(X, obst={"tree": tree}, vart={"tree": tree}, label=None)
    # Toplogy is immutable
    with pytest.raises(nx.NetworkXError):
        tdata.obst["tree"].remove_node("0")
    # Attributes are mutable
    nx.set_node_attributes(tdata.obst["tree"], True, "test")
    assert all(tdata.obst["tree"].nodes[node]["test"] for node in tdata.obst["tree"].nodes)
    # Topology mutable on copy
    tree = tdata.obst["tree"].copy()
    tree.remove_node("1")
    tdata.obst["tree"] = tree
    assert list(tdata.obst["tree"].nodes) == ["root", "0"]


def test_bad_tree(X):
    # Not directed graph
    not_di_graph = nx.Graph()
    with pytest.raises(ValueError):
        _ = td.TreeData(X, obst={"tree": not_di_graph})
    # Has cycle
    has_cycle = nx.DiGraph()
    has_cycle.add_edges_from([("0", "1"), ("1", "0")])
    has_cycle.add_node("2")
    with pytest.raises(ValueError):
        _ = td.TreeData(X, obst={"tree": has_cycle})
    # Not fully connected
    not_fully_connected = nx.DiGraph()
    not_fully_connected.add_edges_from([("root", "0"), ("bad", "1")])
    with pytest.raises(ValueError):
        _ = td.TreeData(X, obst={"tree": not_fully_connected})
    # Leaves not in dim_names
    bad_leaves = nx.DiGraph()
    bad_leaves.add_edges_from([("root", "0"), ("root", "bad")])
    with pytest.raises(ValueError):
        _ = td.TreeData(X, obst={"tree": bad_leaves})
    # Multiple roots
    multi_root = nx.DiGraph()
    multi_root.add_edges_from([("0", "1"), ("1", "0"), ("2", "3")])
    with pytest.raises(ValueError):
        _ = td.TreeData(X, obst={"tree": multi_root})


def test_to_adata(X, tree):
    obs = pd.DataFrame({"cell": ["A", "B", "B"]}, index=["0", "1", "2"])
    tdata = td.TreeData(X, obst={"tree": tree}, obs=obs)
    adata = tdata.to_adata()
    assert type(adata) is ad.AnnData
    assert tdata.X is adata.X
    assert tdata.obs["cell"].tolist() == adata.obs["cell"].tolist()


def test_copy(adata, tree):
    treedata = td.TreeData(adata, obst={"tree": tree})
    treedata_copy = treedata.copy()
    assert np.array_equal(treedata.X, treedata_copy.X)
    assert treedata.obst["tree"].nodes == treedata_copy.obst["tree"].nodes
    assert treedata.obst["tree"].edges == treedata_copy.obst["tree"].edges


def test_transpose(adata, tree):
    treedata = td.TreeData(adata, obst={"tree": tree})
    treedata_transpose = treedata.transpose()
    assert np.array_equal(treedata.X.T, treedata_transpose.X)
    assert treedata.obst["tree"].nodes == treedata_transpose.vart["tree"].nodes
    assert treedata_transpose.obst_keys() == []
    assert np.array_equal(treedata.obs_names, treedata.T.obs_names)


@pytest.mark.parametrize("dim", ["obs", "var"])
def test_not_unique(X, tree, dim):
    with pytest.warns(UserWarning):
        tdata = td.TreeData(pd.DataFrame(X, index=["0", "1", "1"], columns=["0", "1", "1"]))
    assert not getattr(tdata, f"{dim}_names").is_unique
    with pytest.warns(UserWarning):
        setattr(tdata, f"{dim}t", {"tree": tree})
    assert getattr(tdata, f"{dim}_names").is_unique



from collections import deque

import networkx as nx


def subset_tree(tree: nx.DiGraph, leaves: list[str], asview: bool) -> nx.DiGraph:
    """Subset tree."""
    keep_nodes = set(leaves)
    nodes_to_check = deque()
    for node in leaves:
        nodes_to_check.extend(tree.predecessors(node))
    while nodes_to_check:
        node = nodes_to_check.popleft()
        if node in keep_nodes:
            continue
        else:
            keep_nodes.add(node)
            nodes_to_check.extend(tree.predecessors(node))
    if asview:
        return tree.subgraph(keep_nodes)
    else:
        return tree.subgraph(keep_nodes).copy()


def combine_trees(subsets: list[nx.DiGraph]) -> nx.DiGraph:
    """Combine two or more subsets of a tree into a single tree."""
    # Initialize a new directed graph for the combined tree
    combined_tree = nx.DiGraph()

    # Iterate through each subset and add its nodes and edges to the combined tree
    for subset in subsets:
        combined_tree.add_nodes_from(subset.nodes(data=True))
        combined_tree.add_edges_from(subset.edges(data=True))

    # The combined_tree now contains all nodes and edges from the subsets
    return combined_tree


def digraph_to_dict(G: nx.DiGraph) -> dict:
    """Convert a networkx.DiGraph to a dictionary."""
    G = nx.DiGraph(G)
    edge_dict = nx.to_dict_of_dicts(G)
    # Get node data
    node_dict = {node: G.nodes[node] for node in G.nodes()}
    # Combine edge and node data in one dictionary
    graph_dict = {"edges": edge_dict, "nodes": node_dict}

    return graph_dict


def dict_to_digraph(graph_dict: dict) -> nx.DiGraph:
    """Convert a dictionary to a networkx.DiGraph."""
    G = nx.DiGraph()
    # Add nodes and their attributes
    for node, attrs in graph_dict["nodes"].items():
        G.add_node(node, **attrs)
    # Add edges and their attributes
    for source, targets in graph_dict["edges"].items():
        for target, attrs in targets.items():
            G.add_edge(source, target, **attrs)
    return G



"""Code for merging/concatenating TreeData objects."""

from __future__ import annotations

import typing
from collections.abc import (
    Callable,
    Collection,
)
from functools import reduce
from typing import Any, Literal

import anndata as ad
import pandas as pd
from anndata._core.merge import _resolve_dim, resolve_merge_strategy

from treedata._utils import combine_trees

from .treedata import TreeData

StrategiesLiteral = Literal["same", "unique", "first", "only"]


def concat(
    tdatas: Collection[TreeData] | typing.Mapping[str, TreeData],
    *,
    axis: Literal[0, 1] = 0,
    join: Literal["inner", "outer"] = "inner",
    merge: StrategiesLiteral | Callable | None = None,
    uns_merge: StrategiesLiteral | Callable | None = None,
    label: str | None = None,
    keys: Collection | None = None,
    fill_value: Any | None = None,
    pairwise: bool = False,
) -> TreeData:
    """Concatenates TreeData objects along an axis.

    Parameters
    ----------
    tdatas
        The objects to be concatenated. If a Mapping is passed, keys are used for the `keys`
        argument and values are concatenated.
    axis
        Which axis to concatenate along.
    join
        How to align values when concatenating. If "outer", the union of the other axis
        is taken. If "inner", the intersection.
        for more.
    merge
        How elements not aligned to the axis being concatenated along are selected.
        Currently implemented strategies include:

        * `None`: No elements are kept.
        * `"same"`: Elements that are the same in each of the objects.
        * `"unique"`: Elements for which there is only one possible value.
        * `"first"`: The first element seen at each from each position.
        * `"only"`: Elements that show up in only one of the objects.
    uns_merge
        How the elements of `.uns` are selected. Uses the same set of strategies as
        the `merge` argument, except applied recursively.
    label
        Column in axis annotation (i.e. `.obs` or `.var`) to place batch information in.
        If it's None, no column is added.
    keys
        Names for each object being added. These values are used for column values for
        `label` or appended to the index if `index_unique` is not `None`. Defaults to
        incrementing integer labels.
    fill_value
        When `join="outer"`, this is the value that will be used to fill the introduced
        indices. By default, sparse arrays are padded with zeros, while dense arrays and
        DataFrames are padded with missing values.
    pairwise
        Whether pairwise elements along the concatenated dimension should be included.
        This is False by default, since the resulting arrays are often not meaningful.
    """
    axis, dim = _resolve_dim(axis=axis)
    alt_axis, alt_dim = _resolve_dim(axis=1 - axis)
    merge = resolve_merge_strategy(merge)

    # Check indices
    concat_indices = pd.concat([pd.Series(getattr(t, f"{dim}_names")) for t in tdatas], ignore_index=True)
    if not concat_indices.is_unique:
        raise ValueError(f"{dim}_names must be unique to concatenate along axis {axis}")
    alt_indices = [getattr(t, f"{alt_dim}_names") for t in tdatas]
    if join == "inner":
        alt_indices = reduce(lambda x, y: x.intersection(y), alt_indices)
    else:
        alt_indices = reduce(lambda x, y: x.union(y), alt_indices)

    # Concatenate anndata
    adata = ad.concat(
        tdatas,
        axis=axis,
        join=join,
        merge=merge,
        uns_merge=uns_merge,
        label=label,
        keys=keys,
        index_unique=None,
        fill_value=fill_value,
        pairwise=pairwise,
    )
    tdata = TreeData(adata, allow_overlap=True)

    # Trees for concatenation axis
    concat_trees = [getattr(t, f"{dim}t") for t in tdatas]
    unique_keys = {key for mapping in concat_trees for key in mapping.keys()}
    for key in unique_keys:
        trees = [mapping[key] for mapping in concat_trees if key in mapping]
        tree = combine_trees(trees)
        getattr(tdata, f"{dim}t")[key] = tree

    # Trees for other axis
    if join == "inner" and alt_axis == 0:
        tdatas = [t[alt_indices, :] for t in tdatas]
    elif join == "inner" and alt_axis == 1:
        tdatas = [t[:, alt_indices] for t in tdatas]
    alt_trees = merge([getattr(t, f"{alt_dim}t") for t in tdatas])
    for key, tree in alt_trees.items():
        getattr(tdata, f"{alt_dim}t")[key] = tree

    return tdata
